import sys
from typing import List, Union

import pytest

from attrs_strict import type_validator

try:
    from unittest.mock import MagicMock
except ImportError:
    from mock import Mock as MagicMock


@pytest.mark.parametrize(
    "element, type_, error_message",
    [
        (
            2.0,
            Union[int, str],
            "Value of foo 2.0 is not of type typing.Union[int, str]",
        ),
        (
            [1, 2, "p"],
            List[Union[None, int]],
            (
                "Value of foo p is not of type {} in [1, 2, 'p']".format(
                    "typing.Optional[int]"
                    if sys.version_info >= (3, 9)
                    else "typing.Union[NoneType, int]"
                )
            ),
        ),
    ],
)
def test_union_when_type_is_not_specified_raises(element, type_, error_message):

    validator = type_validator()

    attr = MagicMock()
    attr.name = "foo"
    attr.type = type_

    with pytest.raises(ValueError) as error:
        validator(None, attr, element)

    repr_msg = "<{}>".format(error_message)
    assert repr_msg == repr(error.value)


@pytest.mark.parametrize(
    "element, type_,",
    [
        (2.0, Union[int, float]),
        ([1, 2, None, 4, 5], List[Union[None, int]]),
        (None, Union[int, None]),
    ],
)
def test_union_not_raise_for_correct_values(element, type_):
    validator = type_validator()

    attr = MagicMock()
    attr.name = "foo"
    attr.type = type_

    validator(None, attr, element)



import typing
import attr

def resolve_types(
    cls: type,
    global_ns: typing.Optional[typing.Dict[str, typing.Any]] = None,
    local_ns: typing.Optional[typing.Dict[str, typing.Any]] = None,
) -> None: ...
def type_validator(
    empty_ok: bool = True,
) -> typing.Callable[
    [typing.Any, attr.Attribute[typing.Any], typing.Any], None
]:
    def _validator(
        instance: typing.Any,
        attribute: attr.Attribute[typing.Any],
        field: typing.Any,
    ) -> None: ...
    return _validator

def _validate_elements(
    attribute: attr.Attribute[typing.Any],
    value: typing.Any,
    expected_type: typing.Optional[typing.Type[typing.Any]],
) -> None: ...
def _get_base_type(
    type_: typing.Type[typing.Any],
) -> typing.Type[typing.Any]: ...
def _type_matching(
    actual: typing.Type[typing.Any], expected: typing.Type[typing.Any]
) -> bool: ...
def _handle_callable(
    attribute: attr.Attribute[typing.Any],
    callable_: typing.Callable[..., typing.Any],
    expected_type: typing.Type[typing.Callable[..., typing.Any]],
) -> None: ...
def _handle_set_or_list(
    attribute: attr.Attribute[typing.Any],
    container: typing.Union[typing.Set[typing.Any], typing.List[typing.Any]],
    expected_type: typing.Union[
        typing.Type[typing.Set[typing.Any]],
        typing.Type[typing.List[typing.Any]],
    ],
) -> None: ...
def _handle_dict(
    attribute: attr.Attribute[typing.Any],
    container: typing.Union[
        typing.Mapping[typing.Any, typing.Any],
        typing.MutableMapping[typing.Any, typing.Any],
    ],
    expected_type: typing.Union[
        typing.Type[typing.Mapping[typing.Any, typing.Any]],
        typing.Type[typing.MutableMapping[typing.Any, typing.Any]],
    ],
) -> None: ...
def _handle_tuple(
    attribute: attr.Attribute[typing.Any],
    container: typing.Tuple[typing.Any],
    expected_type: typing.Type[typing.Tuple[typing.Any]],
) -> None: ...
def _handle_union(
    attribute: attr.Attribute[typing.Any],
    value: typing.Any,
    expected_type: typing.Type[typing.Any],
) -> None: ...



from ._error import (
    AttributeTypeError as AttributeTypeError,
    BadTypeError as BadTypeError,
    TupleError as TupleError,
    TypeValidationError as TypeValidationError,
    UnionError as UnionError,
)
from ._type_validation import type_validator as type_validator



import typing
import attr
import inspect

class TypeValidationError(Exception):
    def __repr__(self) -> str: ...

class BadTypeError(TypeValidationError, ValueError):
    def __init__(self) -> None:
        self.containers: typing.List[typing.Iterable[typing.Any]]
    def add_container(self, container: typing.Any) -> None: ...
    def _render(self, error: str) -> str: ...

class AttributeTypeError(BadTypeError):
    def __init__(
        self, value: typing.Any, attribute: attr.Attribute[typing.Any]
    ) -> None: ...
    def __str__(self) -> str: ...

class CallableError(BadTypeError):
    def __init__(
        self,
        attribute: attr.Attribute[typing.Any],
        callable_signature: inspect.Signature,
        expected_signature: typing.Type[typing.Callable[..., typing.Any]],
        mismatch_callable_arg: inspect.Parameter,
        expected_callable_arg: inspect.Parameter,
    ) -> None: ...
    def __str__(self) -> str: ...

class EmptyError(BadTypeError):
    def __init__(
        self, container: typing.Any, attribute: attr.Attribute[typing.Any]
    ) -> None: ...
    def __str__(self) -> str: ...

class TupleError(BadTypeError):
    def __init__(
        self,
        container: typing.Any,
        attribute: typing.Optional[typing.Type[typing.Any]],
        tuple_types: typing.Tuple[typing.Type[typing.Any]],
    ) -> None: ...
    def __str__(self) -> str: ...
    def _more_or_less(self) -> str: ...

class UnionError(BadTypeError):
    def __init__(
        self,
        container: typing.Any,
        attribute: str,
        expected_type: typing.Type[typing.Any],
    ) -> None: ...
    def __str__(self) -> str: ...



from typing import Tuple

import pytest

from attrs_strict import type_validator

try:
    from unittest.mock import MagicMock
except ImportError:
    from mock import Mock as MagicMock


def test_tuple_with_incorrect_number_of_arguments_raises():
    element = (1, 2, 3, 4)

    attr = MagicMock()
    attr.name = "foo"
    attr.type = Tuple[int, int, int]

    validator = type_validator()

    with pytest.raises(ValueError) as error:
        validator(None, attr, element)

    assert (
        "<Element (1, 2, 3, 4) has more elements than types specified "
        "in typing.Tuple[int, int, int]. Expected 3 received 4>"
    ) == repr(error.value)


def test_tuple_of_tuple_raises():
    element = ((1, 2), (3, 4, 5))

    attr = MagicMock()
    attr.name = "foo"
    attr.type = Tuple[Tuple[int, int], Tuple[int, int]]

    validator = type_validator()

    with pytest.raises(ValueError) as error:
        validator(None, attr, element)

    assert (
        "<Element (3, 4, 5) has more elements than types specified "
        "in typing.Tuple[typing.Tuple[int, int], typing.Tuple[int, int]]. "
        "Expected 2 received 3 in ((1, 2), (3, 4, 5))>"
    ) == repr(error.value)


def test_variable_length_tuple():
    element = (1, 2, 3, 4)

    attr = MagicMock()
    attr.name = "foo"
    attr.type = Tuple[int, ...]

    validator = type_validator()

    validator(None, attr, element)


def test_variable_length_tuple_empty():
    element = ()

    attr = MagicMock()
    attr.name = "foo"
    attr.type = Tuple[int, ...]

    validator = type_validator()

    validator(None, attr, element)


def test_variable_length_tuple_raises():
    element = (1, 2, 3, "4")

    attr = MagicMock()
    attr.name = "foo"
    attr.type = Tuple[int, ...]

    validator = type_validator()

    with pytest.raises(ValueError) as error:
        validator(None, attr, element)

    assert (
        "<foo must be typing.Tuple[int, ...] (got 4 that is a {}) "
        "in (1, 2, 3, '4')>"
    ).format(str) == repr(error.value)



import typing

import pytest

from attrs_strict import AttributeTypeError, BadTypeError, type_validator

try:
    from unittest.mock import MagicMock
except ImportError:
    from mock import Mock as MagicMock


@pytest.mark.parametrize("test_type, correct", [(str, "foo"), (int, 42)])
def test_typing_newtype_single_validation_success(test_type, correct):
    SomeNew = typing.NewType("SomeNew", test_type)

    validator = type_validator()
    attr = MagicMock()
    attr.type = SomeNew

    validator(None, attr, correct)
    validator(None, attr, SomeNew(correct))


@pytest.mark.parametrize(
    "test_type, wrongs", [(str, [42, True]), (int, ["foo", ()])]
)
def test_typing_newtype_single_validation_failure(test_type, wrongs):
    SomeNew = typing.NewType("SomeNew", test_type)

    validator = type_validator()
    attr = MagicMock()
    attr.type = SomeNew

    for wrong in wrongs:
        with pytest.raises(AttributeTypeError) as error:
            validator(None, attr, wrong)

    assert "must be NewType(SomeNew, {})".format(str(test_type)) in str(
        error.value
    )


@pytest.mark.parametrize(
    "container, test_type, correct",
    [
        (typing.List, str, ["foo", "bar"]),
        (typing.Tuple, int, (0,)),
        (typing.Optional, str, None),
    ],
)
def test_typing_newtype_within_container_validation_success(
    container, test_type, correct
):
    SomeNew = typing.NewType("SomeNew", test_type)

    validator = type_validator()
    attr = MagicMock()
    attr.type = container[SomeNew]

    validator(None, attr, correct)


@pytest.mark.parametrize(
    "container, test_type, wrongs",
    [
        (typing.List, str, [42, True, "foo", ("foo", "bar")]),
        (typing.Tuple, int, ["foo", 42, [0, 1, "2"]]),
        (typing.Optional, str, [42, (1, 2)]),
    ],
)
def test_typing_newtype_within_container_validation_failure(
    container, test_type, wrongs
):
    SomeNew = typing.NewType("SomeNew", test_type)

    validator = type_validator()
    attr = MagicMock()
    attr.type = container[SomeNew]

    for wrong in wrongs:
        with pytest.raises(BadTypeError) as error:
            validator(None, attr, wrong)

    assert "must be {}".format(str(attr.type)) in str(
        error.value
    ) or "is not of type {}".format(str(attr.type)) in str(error.value)


# -----------------------------------------------------------------------------
# Copyright 2019 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------- END-OF-FILE -----------------------------------



import collections
from typing import Any, DefaultDict, Dict, List, Mapping, MutableMapping

import pytest

from attrs_strict import type_validator

try:
    from collections.abc import Mapping as CollectionsMapping
    from collections.abc import MutableMapping as CollectionsMutableMapping
except ImportError:
    from collections import Mapping as CollectionsMapping
    from collections import MutableMapping as CollectionsMutableMapping


try:
    from unittest.mock import MagicMock
except ImportError:
    from mock import Mock as MagicMock


def test_defaultdict_raise_error():
    elem = collections.defaultdict(int)
    elem[5] = [1, 2, 3]

    validator = type_validator()

    attr = MagicMock()
    attr.name = "foo"
    attr.type = DefaultDict[int, List[str]]

    with pytest.raises(ValueError) as error:
        validator(None, attr, elem)

    assert (
        "<foo must be typing.DefaultDict[int, typing.List[str]] "
        "(got 1 that is a {}) in [1, 2, 3] in "
        "defaultdict({}, {{5: [1, 2, 3]}})>"
    ).format(int, int) == repr(error.value)


def test_defaultdict_with_correct_type_no_raise():
    elem = collections.defaultdict(int)
    elem[5] = [1, 2, 3]
    elem[6] = [4, 5, 6]

    validator = type_validator()

    attr = MagicMock()
    attr.name = "foo"
    attr.type = DefaultDict[int, List[int]]

    validator(None, attr, elem)


def test_dict_with_any_does_not_raise():
    elem = {"foo": 123, "b": "abc"}

    validator = type_validator()

    attr = MagicMock()
    attr.name = "zoo"
    attr.type = Dict[str, Any]

    validator(None, attr, elem)


@pytest.mark.parametrize(
    "data, type, validator_type, error_message",
    [
        (
            {"foo": 123},
            CollectionsMapping,
            Mapping,
            (
                "<zoo must be typing.Mapping[str, str] "
                "(got 123 that is a {})"
            ).format(int),
        ),
        (
            {1: "boo", 2: "zoo"},
            CollectionsMutableMapping,
            MutableMapping,
            (
                "<zoo must be typing.MutableMapping[str, str] "
                "(got 1 that is a {})"
            ).format(int),
        ),
    ],
)
def test_abc_mapping_types_throw_when_type_is_wrong(
    data, type, validator_type, error_message
):
    class TestMapping(type):
        def __init__(self, items):
            self._data = items

        def __getitem__(self, item):
            return self._data[item]

        def __len__(self):
            return len(self._data)

        def __iter__(self):
            return iter(self._data)

        def __delitem__(self, item):
            pass

        def __setitem__(self, item, value):
            pass

    validator = type_validator()

    attr = MagicMock()
    attr.name = "zoo"
    attr.type = validator_type[str, str]

    with pytest.raises(ValueError) as error:
        validator(None, attr, TestMapping(data))

    assert error_message in repr(error.value)



import sys
import typing

import attr
import pytest

from attrs_strict import AttributeTypeError, CallableError, type_validator

not_on_py2 = pytest.mark.xfail(
    sys.version_info < (3, 5),
    reason="Type annotations for callables are not supported in py2",
)


class _TestResources:

    NewInt = typing.NewType("NewType", int)

    def plain_unannotated_callable(a, b):
        pass

    def int_int_returns_str(a, b):
        pass

    int_int_returns_str.__annotations__ = {"a": int, "b": int, "return": str}

    def newint_int_returns_str(a, b):
        pass

    newint_int_returns_str.__annotations__ = {
        "a": NewInt,
        "b": int,
        "return": str,
    }

    def int_int_returns_int(a, b):
        pass

    int_int_returns_int.__annotations__ = {"a": int, "b": int, "return": int}

    def dict_str_int_returns_int(a):
        pass

    dict_str_int_returns_int.__annotations__ = {
        "a": typing.Dict[str, int],
        "return": int,
    }

    def int_returns_int(a):
        pass

    int_returns_int.__annotations__ = {"a": int, "return": int}

    def dict_int_str_returns_int(a):
        pass

    dict_int_str_returns_int.__annotations__ = {
        "a": typing.Dict[int, str],
        "return": int,
    }

    def dict_str_str_returns_int(a):
        pass

    dict_str_str_returns_int.__annotations__ = {
        "a": typing.Dict[str, str],
        "return": int,
    }

    def int_default_returns_int(a=123):
        pass

    int_default_returns_int.__annotations__ = {"a": int, "return": int}


@not_on_py2
@pytest.mark.parametrize(
    "name, callable_, attribute_type",
    [
        (
            "no-types-defined-matches-typed-callable",
            _TestResources.int_int_returns_str,
            typing.Callable,
        ),
        (
            "no-types-defined-matches-untyped-callable",
            _TestResources.plain_unannotated_callable,
            typing.Callable,
        ),
        (
            "list_of_untyped_callables",
            [_TestResources.plain_unannotated_callable] * 3,
            typing.List[typing.Callable],
        ),
        (
            "typed_callable_expected_has_newtype",
            _TestResources.int_int_returns_str,
            typing.Callable[[_TestResources.NewInt, int], str],
        ),
        (
            "typed_callable_value_has_newtype",
            _TestResources.newint_int_returns_str,
            typing.Callable[[int, int], str],
        ),
        (
            "typed_callable_",
            _TestResources.int_int_returns_str,
            typing.Callable[[int, int], str],
        ),
        (
            "list_of_typed_callables",
            [_TestResources.int_int_returns_str] * 3,
            typing.List[typing.Callable[[int, int], str]],
        ),
        (
            "union_type_arg2_provided",
            _TestResources.int_int_returns_str,
            typing.Callable[[int, int], typing.Union[int, str]],
        ),
        (
            "union_type_arg1_provided",
            _TestResources.int_int_returns_int,
            typing.Callable[[int, int], typing.Union[int, str]],
        ),
        (
            "dict_type",
            _TestResources.dict_str_int_returns_int,
            typing.Callable[[typing.Dict[str, int]], int],
        ),
        (
            "union_of_int_dict_arg1_provided",
            _TestResources.int_returns_int,
            typing.Callable[[typing.Union[int, typing.Dict[int, str]]], int],
        ),
        (
            "union_of_int_dict_arg2_provided",
            _TestResources.dict_int_str_returns_int,
            typing.Callable[[typing.Union[int, typing.Dict[int, str]]], int],
        ),
        (
            "nested_unions_arg1_provided",
            _TestResources.int_returns_int,
            typing.Callable[
                [typing.Union[int, typing.Dict[typing.Union[str, int], str]]],
                int,
            ],
        ),
        (
            "nested_unions_arg2_provided",
            _TestResources.dict_str_str_returns_int,
            typing.Callable[
                [typing.Union[int, typing.Dict[typing.Union[str, int], str]]],
                int,
            ],
        ),
        (
            "any_type_int_provided",
            _TestResources.int_int_returns_int,
            typing.Callable[[int, int], typing.Any],
        ),
        (
            "any_type_str_provided",
            _TestResources.int_int_returns_str,
            typing.Callable[[int, int], typing.Any],
        ),
        (
            "callable_with_default_arg",
            _TestResources.int_default_returns_int,
            typing.Callable[[int], int],
        ),
    ],
)
def test_callable_not_raises_with_valid_annotations(
    name, callable_, attribute_type
):
    @attr.s
    class Something(object):
        call_me = attr.ib(validator=type_validator(), type=attribute_type)

    Something(call_me=callable_)


@not_on_py2
@pytest.mark.parametrize(
    "name, callable_, attribute_type, raised_error_type",
    [
        (
            "not_a_callable",
            "clearly-not-a-callable",
            typing.Callable,
            AttributeTypeError,
        ),
        (
            "callable_with_incorrect_types",
            _TestResources.int_int_returns_int,
            typing.Callable[[int, int], str],
            CallableError,
        ),
        (
            "list_of_callables_different_types",
            [
                _TestResources.int_int_returns_int,
                _TestResources.int_int_returns_str,
            ],
            typing.List[typing.Callable[[int, int], str]],
            CallableError,
        ),
    ],
)
def test_callable_raises_with_invalid_types(
    name, callable_, attribute_type, raised_error_type
):
    @attr.s
    class Something(object):
        call_me = attr.ib(validator=type_validator(), type=attribute_type)

    with pytest.raises(raised_error_type):
        Something(call_me=callable_)



"""Runtime validation library for attrs data classes.

"""

from ._error import (
    AttributeTypeError,
    BadTypeError,
    CallableError,
    TupleError,
    TypeValidationError,
    UnionError,
)
from ._type_validation import type_validator
from ._version import __version__  # noqa

__all__ = [
    "type_validator",
    "AttributeTypeError",
    "BadTypeError",
    "UnionError",
    "TupleError",
    "TypeValidationError",
    "CallableError",
]

# -----------------------------------------------------------------------------
# Copyright 2019 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------- END-OF-FILE -----------------------------------



def is_newtype(type_):
    return (
        hasattr(type_, "__name__")
        and hasattr(type_, "__supertype__")
        and type_.__module__ == "typing"
    )


def format_type(type_):
    if is_newtype(type_):
        return "NewType({}, {})".format(type_.__name__, type_.__supertype__)

    return str(type_)



import typing

import attr
import pytest

from attrs_strict import type_validator

try:
    from unittest.mock import MagicMock
except ImportError:
    from mock import Mock as MagicMock


@pytest.mark.parametrize(
    "items, types, message",
    [
        (
            [1, 2, 3],
            typing.Set,
            "Smth must be {} (got [1, 2, 3] that is a {})".format(
                typing.Set, list
            ),
        ),
        (
            [1, 2, 3],
            typing.Dict,
            "Smth must be {} (got [1, 2, 3] that is a {})".format(
                typing.Dict, list
            ),
        ),
        (
            [1, 2, 3],
            typing.Tuple,
            "Smth must be {} (got [1, 2, 3] that is a {})".format(
                typing.Tuple, list
            ),
        ),
    ],
)
def test_container_is_not_of_expected_type_raises_TypeError(
    items, types, message
):
    validator = type_validator()

    attr = MagicMock()
    attr.name = "Smth"
    attr.type = types

    with pytest.raises(ValueError) as error:
        validator(None, attr, items)

    repr_msg = "<{}>".format(message)
    assert repr_msg == repr(error.value)


def test_does_not_raise_when_container_is_empty_and_allowed():
    items = []

    validator = type_validator(True)
    attr = MagicMock()
    attr.type = typing.List[int]

    validator(None, attr, items)


def test_raises_when_container_is_empty_and_empty_ok_is_false():
    # GIVEN
    items = []
    validator = type_validator(empty_ok=False)
    attr = MagicMock()
    attr.name = "Smth"
    attr.type = str

    # WHEN
    with pytest.raises(ValueError) as error:
        validator(None, attr, items)

    # THEN
    assert "Smth can not be empty and must be str"
    "(got None)" == str(error.value)


def test_no_type_specified_is_fine():
    @attr.s
    class Something(object):
        numbers = attr.ib(validator=type_validator())

    Something([1, 2, 3, 4])


# -----------------------------------------------------------------------------
# Copyright 2019 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------- END-OF-FILE -----------------------------------



from typing import Any, Dict, List, Set, Tuple

import pytest

from attrs_strict import type_validator

try:
    from unittest.mock import MagicMock
except ImportError:
    from mock import Mock as MagicMock


@pytest.mark.parametrize(
    "values, type_, error_message",
    [
        (
            [1, 2, "a"],
            List[int],
            (
                (
                    "numbers must be typing.List[int] "
                    "(got a that is a {}) in [1, 2, 'a']"
                ).format(str)
            ),
        ),
        (
            [[1, "a"]],
            List[List[int]],
            (
                (
                    "numbers must be "
                    "typing.List[typing.List[int]] (got a that is a {}) "
                    "in [1, 'a'] in [[1, 'a']]"
                ).format(str)
            ),
        ),
        (
            [[1, 2, 3], ["a"]],
            List[List[int]],
            (
                (
                    "numbers must be "
                    "typing.List[typing.List[int]] (got a that is a {}) "
                    "in ['a'] in [[1, 2, 3], ['a']]"
                ).format(str)
            ),
        ),
        (
            [(1, 2, "foo")],
            List[Tuple[int, int, int]],
            (
                "numbers must be "
                "typing.List[typing.Tuple[int, int, int]] (got foo "
                "that is a {}) in (1, 2, 'foo') in "
                "[(1, 2, 'foo')]"
            ).format(str),
        ),
    ],
)
def test_list_of_values_raise_value_error(values, type_, error_message):

    validator = type_validator()

    attrib = MagicMock()
    attrib.name = "numbers"
    attrib.type = type_

    with pytest.raises(ValueError) as error:
        validator(None, attrib, values)

    # THEN
    msg = "<{}>".format(error_message)
    assert msg == repr(error.value)


@pytest.mark.parametrize(
    "values, type_",
    [
        ([1, 2, 3], List[int]),
        ([[1], [2], [3]], List[List[int]]),
        ({1, 2, 3}, Set[int]),
        ([{1: [1, 2, 3], 2: [3, 4, 5]}], List[Dict[int, List[int]]]),
        ([1, 2, 3, 4], List[Any]),
        ([1, 2, {"foo": "bar"}], List[Any]),
    ],
)
def test_list_of_valid_values_no_raise(values, type_):
    validator = type_validator()

    attrib = MagicMock()
    attrib.name = "numbers"
    attrib.type = type_

    validator(None, attrib, values)



import attr
import pytest

from attrs_strict import type_validator


@pytest.mark.parametrize(
    "value, expected, actual",
    [
        (3, str, int),
        ("five", int, str),
        (None, str, type(None)),
        (2.3, int, float),
    ],
)
def test_primitive_types(value, expected, actual):
    @attr.s
    class Something(object):
        number = attr.ib(validator=type_validator(), type=expected)

    with pytest.raises(ValueError) as error:
        Something(number=value)

    assert repr(
        error.value
    ) == "<number must be {} (got {} that is a {})>".format(
        expected, value, actual
    )


def test_reassign_evaluate():
    @attr.s
    class Something(object):
        number = attr.ib(validator=type_validator(), type=str)

    x = Something(number="foo")
    with pytest.raises(ValueError) as error:
        x.number = 5
        attr.validate(x)

    assert repr(
        error.value
    ) == "<number must be {} (got 5 that is a {})>".format(str, int)



import os
import sys

from attrs_strict import __version__

sys.path.insert(0, os.path.abspath("../.."))

project = ""
copyright = "2019, Bloomberg"
author = "Erik-Cristian Seulean"

release = "2019"

extensions = [
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
]

templates_path = ["_templates"]
language = "python"

exclude_patterns = []

html_theme = "alabaster"
html_static_path = []

master_doc = "index"

version = u".".join(__version__.split(".")[:2])
release = __version__



from ._commons import format_type


class TypeValidationError(Exception):
    def __repr__(self):
        return "<{}>".format(str(self))


class BadTypeError(TypeValidationError, ValueError):
    def __init__(self):
        self.containers = []

    def add_container(self, container):
        self.containers.append(container)

    def _render(self, error):
        if self.containers:
            backtrack = " in ".join(
                [str(container) for container in self.containers]
            )
            return "{} in {}".format(error, backtrack)

        return error


class AttributeTypeError(BadTypeError):
    def __init__(self, value, attribute):
        super(AttributeTypeError, self).__init__()
        self.value = value
        self.attribute = attribute

    def __str__(self):
        if self.attribute.type is None:
            error = (
                "attrs-strict error: AttributeTypeError was "
                "raised on an attribute ({}) with no defined type".format(
                    self.attribute.name
                )
            )
        else:
            error = "{} must be {} (got {} that is a {})".format(
                self.attribute.name,
                format_type(self.attribute.type),
                self.value,
                type(self.value),
            )

        return self._render(error)


class CallableError(BadTypeError):
    def __init__(
        self,
        attribute,
        callable_signature,
        expected_signature,
        mismatch_callable_arg,
        expected_callable_arg,
    ):
        super(CallableError, self).__init__()
        self.attribute = attribute
        self.callable_signature = callable_signature
        self.expected_signature = expected_signature
        self.mismatch_callable_arg = mismatch_callable_arg
        self.expected_callable_arg = expected_callable_arg

    def __str__(self):
        error_msg = (
            "Error with: {} . Expected Callable "
            "signature: {} got: {}. {} should be {}"
        ).format(
            self.attribute.name,
            self.expected_signature,
            self.callable_signature,
            self.mismatch_callable_arg,
            self.expected_callable_arg,
        )
        return self._render(error_msg)


class EmptyError(BadTypeError):
    def __init__(self, container, attribute):
        super(EmptyError, self).__init__()
        self.container = container
        self.attribute = attribute

    def __str__(self):
        if self.attribute.type is None:
            error = (
                "attrs-strict error: AttributeTypeError was "
                "raised on an attribute ({}) with no defined type".format(
                    self.attribute.name
                )
            )
        else:
            error = "{} can not be empty and must be {} (got {})".format(
                self.attribute.name,
                format_type(self.attribute.type),
                self.container,
            )

        return self._render(error)


class TupleError(BadTypeError):
    def __init__(self, container, attribute, tuple_types):
        super(TupleError, self).__init__()
        self.attribute = attribute
        self.container = container
        self.tuple_types = tuple_types

    def __str__(self):
        error = (
            "Element {} has {} elements than types "
            "specified in {}. Expected {} received {}"
        ).format(
            self.container,
            self._more_or_less(),
            self.attribute,
            len(self.tuple_types),
            len(self.container),
        )

        return self._render(error)

    def _more_or_less(self):
        return "more" if len(self.container) > len(self.tuple_types) else "less"


class UnionError(BadTypeError):
    def __init__(self, container, attribute, expected_type):
        super(UnionError, self).__init__()
        self.attribute = attribute
        self.container = container
        self.expected_type = expected_type

    def __str__(self):
        error = "Value of {} {} is not of type {}".format(
            self.attribute, self.container, self.expected_type
        )

        return self._render(error)



from typing import Optional

import attr
import pytest

from attrs_strict import type_validator


@pytest.mark.parametrize(
    "type_, good_value, bad_value",
    [(str, "str", 0xBAD), (int, 7, "bad"), (Optional[str], None, 0xBAD)],
)
def test_real_types(type_, good_value, bad_value):
    @attr.s(auto_attribs=True)
    class Something:
        value: type_ = attr.ib(validator=type_validator())

    with pytest.raises(ValueError):
        Something(bad_value)

    x = Something(good_value)
    with pytest.raises(ValueError):
        x.value = bad_value
        attr.validate(x)


@attr.s(auto_attribs=True)
class Child:
    parent: "Parent" = attr.ib(validator=type_validator())


@attr.s(auto_attribs=True)
class Parent:
    pass


def test_forward_ref():
    Child(Parent())

    with pytest.raises(ValueError):
        Child(15)


@attr.s(auto_attribs=True)
class Self:
    parent: Optional["Self"] = attr.ib(None, validator=type_validator())


def test_recursive():
    Self(Self())
    Self(Self(None))

    with pytest.raises(ValueError):
        Self(Self(17))



from setuptools import setup

with open("README.md") as fp:
    readme = fp.read()

long_description = "".join(
    [
        section.split("<!-- end -->")[0]
        for section in readme.split("<!-- begin -->")
        if "<!-- end -->" in section
    ]
)

setup(long_description=long_description)



import collections
import typing

import attr

from ._commons import is_newtype
from ._error import (
    AttributeTypeError,
    BadTypeError,
    CallableError,
    EmptyError,
    TupleError,
    UnionError,
)

try:
    from collections.abc import Callable, Mapping, MutableMapping
except ImportError:
    from collections import Callable, Mapping, MutableMapping

try:
    from inspect import signature
except ImportError:
    # silencing type error so mypy doesn't complain about duplicate import
    from funcsigs import signature  # type: ignore

try:
    from itertools import zip_longest
except ImportError:
    # silencing type error so mypy doesn't complain about duplicate import
    from itertools import izip_longest as zip_longest  # type: ignore

try:
    from typing import ForwardRef  # type: ignore # Not in stubs
except ImportError:
    from typing import _ForwardRef as ForwardRef  # type: ignore # Not in stubs


class _StringAnnotationError(Exception):
    """Raised when we find string annotations in a class."""


class SimilarTypes:
    Dict = {
        dict,
        collections.OrderedDict,
        collections.defaultdict,
        Mapping,
        MutableMapping,
        typing.Dict,
        typing.DefaultDict,
        typing.Mapping,
        typing.MutableMapping,
    }
    List = {set, list, typing.List, typing.Set}
    Tuple = {tuple, typing.Tuple}
    Callable = {typing.Callable, Callable}


def resolve_types(cls, global_ns=None, local_ns=None):
    """
    Resolve any strings and forward annotations in type annotations.

    :param type cls: Class to resolve.
    :param globalns: Dictionary containing global variables, if needed.
    :param localns: Dictionary containing local variables, if needed.
    :raise TypeError: If *cls* is not a class.
    :raise attr.exceptions.NotAnAttrsClassError: If *cls* is not an ``attrs``
           class.
    :raise NameError: If types cannot be resolved because of missing variables.

    """
    hints = typing.get_type_hints(cls, globalns=global_ns, localns=local_ns)
    for field in attr.fields(cls):
        if field.name in hints:
            # Since fields have been frozen we must work around it.
            object.__setattr__(field, "type", hints[field.name])


def type_validator(empty_ok=True):
    """
    Validates the attributes using the type argument specified. If the
    type argument is not present, the attribute is considered valid.

    :param empty_ok: Boolean flag that indicates if the field can be empty
                     in case of a collection or None for builtin types.

    """

    def _validator(instance, attribute, field):
        if not empty_ok and not field:
            raise EmptyError(field, attribute)

        try:
            _validate_elements(attribute, field, attribute.type)
        except _StringAnnotationError:
            resolve_types(type(instance))
            _validate_elements(attribute, field, attribute.type)

    return _validator


def _validate_elements(attribute, value, expected_type):
    if expected_type is None:
        return

    base_type = _get_base_type(expected_type)

    if base_type == typing.Any:
        return

    if isinstance(base_type, (str, ForwardRef)):
        # These base_types happen when you have string annotations and cannot
        # be used in isinstance.
        raise _StringAnnotationError()

    if base_type != typing.Union and not isinstance(  # type: ignore
        value, base_type
    ):
        raise AttributeTypeError(value, attribute)

    if base_type == typing.Union:  # type: ignore
        _handle_union(attribute, value, expected_type)
    elif base_type in SimilarTypes.List:
        _handle_set_or_list(attribute, value, expected_type)
    elif base_type in SimilarTypes.Dict:
        _handle_dict(attribute, value, expected_type)
    elif base_type in SimilarTypes.Tuple:
        _handle_tuple(attribute, value, expected_type)
    elif base_type in SimilarTypes.Callable:  # type: ignore
        _handle_callable(attribute, value, expected_type)


def _get_base_type(type_):
    if hasattr(type_, "__origin__") and type_.__origin__ is not None:
        base_type = type_.__origin__  # type: typing.Type[typing.Any]
    elif is_newtype(type_):
        base_type = type_.__supertype__
    else:
        base_type = type_

    return base_type


def _type_matching(actual, expected):
    actual = actual.__supertype__ if is_newtype(actual) else actual
    expected = expected.__supertype__ if is_newtype(expected) else expected

    if expected == actual or expected == typing.Any:
        return True

    base_type = _get_base_type(expected)

    if base_type == typing.Union:  # type: ignore
        return any(
            _type_matching(actual, expected_candidate)
            for expected_candidate in expected.__args__
        )

    elif base_type in (
        SimilarTypes.Dict
        | SimilarTypes.List
        | SimilarTypes.Tuple
        | SimilarTypes.Callable
    ):
        return all(
            _type_matching(actual, expected)
            for actual, expected in zip_longest(
                actual.__args__, expected.__args__
            )
        )

    return False


def _handle_callable(attribute, callable_, expected_type):
    _signature = signature(callable_)
    callable_args = [
        param.annotation for param in _signature.parameters.values()
    ]
    callable_args.append(_signature.return_annotation)
    if not getattr(expected_type, "__args__", None):  # type: ignore
        return  # No annotations specified on type, matches all Callables

    for callable_arg, expected_arg in zip_longest(
        callable_args, expected_type.__args__  # type: ignore
    ):
        if not _type_matching(callable_arg, expected_arg):
            raise CallableError(
                attribute, _signature, expected_type, callable_arg, expected_arg
            )


def _handle_set_or_list(attribute, container, expected_type):
    (element_type,) = expected_type.__args__  # type: ignore

    for element in container:
        try:
            _validate_elements(attribute, element, element_type)
        except BadTypeError as error:
            error.add_container(container)
            raise error


def _handle_dict(attribute, container, expected_type):
    key_type, value_type = expected_type.__args__  # type: ignore

    for key in container:
        try:
            _validate_elements(attribute, key, key_type)
            _validate_elements(attribute, container[key], value_type)
        except BadTypeError as error:
            error.add_container(container)
            raise error


def _handle_tuple(attribute, container, expected_type):
    tuple_types = expected_type.__args__  # type: ignore
    if len(tuple_types) == 2 and tuple_types[1] == Ellipsis:
        element_type = tuple_types[0]
        tuple_types = (element_type,) * len(container)

    if len(container) != len(tuple_types):
        raise TupleError(container, attribute.type, tuple_types)

    for element, expected_type in zip(container, tuple_types):
        try:
            _validate_elements(attribute, element, expected_type)
        except BadTypeError as error:
            error.add_container(container)
            raise error


def _handle_union(attribute, value, expected_type):
    union_has_none_type = any(
        elem is None.__class__ for elem in expected_type.__args__
    )

    if value is None and union_has_none_type:
        return

    for arg in expected_type.__args__:
        try:
            _validate_elements(attribute, value, arg)
            return
        except ValueError:
            pass
    raise UnionError(value, attribute.name, expected_type)


# -----------------------------------------------------------------------------
# Copyright 2019 Bloomberg Finance L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------- END-OF-FILE -----------------------------------
